@@ -1,6 +1,6 @@
 {-# LANGUAGE GADTs, TypeFamilies, TypeOperators, ScopedTypeVariables, CPP #-}
 {-# LANGUAGE StandaloneDeriving, FlexibleInstances #-} 
-{-# LANGUAGE DefaultSignatures, FlexibleContexts, EmptyCase, LambdaCase #-}
+{-# LANGUAGE DefaultSignatures, FlexibleContexts, LambdaCase #-}
 {-# OPTIONS_GHC -Wall -fenable-rewrite-rules #-}
 
 -- ScopedTypeVariables works around a 6.10 bug.  The forall keyword is
@@ -42,7 +42,7 @@
 -- see @examples/Generic.hs@, which can be run with: 
 -- 
 -- @
--- $ cabal configure -fexamples && cabal run generic
+-- cabal configure -fexamples && cabal run generic
 -- @ 
 -- 
 -- 
@@ -55,21 +55,29 @@ module Data.MemoTrie
   , memo, memo2, memo3, mup
   , inTrie, inTrie2, inTrie3
   -- , untrieBits
-  , trieGeneric, untrieGeneric, enumerateGeneric, Reg   
+  , trieGeneric, untrieGeneric, enumerateGeneric, Reg
+  , memoFix
   ) where
 
 -- Export the parts of HasTrie separately in order to get the associated data
 -- type constructors, so I can define instances of other classes on them.
 
+import Data.Function (fix)
 import Data.Bits
 import Data.Word
 import Data.Int
+#if !MIN_VERSION_base(4,8,0)
 import Control.Applicative
+#endif
 import Control.Arrow (first,(&&&))
+#if !MIN_VERSION_base(4,8,0)
 import Data.Monoid
+#endif
 import Data.Function (on)
 import GHC.Generics
 
+import Control.Newtype
+
 import Data.Void (Void) 
  
 -- import Prelude hiding (id,(.))
@@ -128,7 +136,13 @@ untrie3 tt = untrie2 . untrie tt
 -}
 
 
-{-# RULES "trie/untrie"   forall t. trie (untrie t) = t #-}
+-- {-# RULES "trie/untrie"   forall t. trie (untrie t) = t #-}
+
+--     warning: [-Winline-rule-shadowing] …
+--     Rule "trie/untrie" may never fire
+--       because rule "Class op untrie" for ‘untrie’ might fire first
+--     Probable fix: add phase [n] or [~n] to the competing rule
+
 
 -- Don't include the dual rule:
 --   "untrie/trie"   forall f. untrie (trie f) = f
@@ -156,6 +170,42 @@ mup mem f = memo (mem . f)
 memo2 = mup memo
 memo3 = mup memo2
 
+-- | Memoizing recursion. Use like `fix`.
+memoFix :: HasTrie a => ((a -> b) -> (a -> b)) -> (a -> b)
+memoFix h = fix (memo . h)
+
+#if 0
+-- Equivalently,
+
+memoFix h = fix (\ f' -> memo (h f'))
+
+memoFix h = f'
+  where f' = memo (h f')
+
+memoFix h = f'
+ where
+   f' = memo f
+   f  = h f'
+#endif
+
+#if 0
+-- Example
+
+fibF :: (Integer -> Integer) -> (Integer -> Integer)
+fibF _ 0 = 1
+fibF _ 1 = 1
+fibF f n = f (n-1) + f (n-2)
+
+fib :: Integer -> Integer
+fib = fix fibF
+
+fib' :: Integer -> Integer
+fib' = memoFix fibF
+
+-- Try fib 30 vs fib' 30
+#endif
+
+
 -- | Apply a unary function inside of a trie
 inTrie :: (HasTrie a, HasTrie c) =>
           ((a  ->  b) -> (c  ->  d))
@@ -179,17 +229,28 @@ inTrie3 = untrie ~> inTrie2
 
 instance HasTrie Void where
   -- As suggested by Audun Skaugen
-  data Void :->: a = VoidTrie 
+  data Void :->: a = VoidTrie
   trie _ = VoidTrie
-  untrie VoidTrie = \case 
+  untrie VoidTrie = \ _ -> error "untrie VoidTrie"
+                    -- \case  -- needs EmptyCase
   enumerate VoidTrie = []
 
+instance Newtype (Void :->: a) where
+  type O (Void :->: a) = ()
+  pack () = VoidTrie
+  unpack VoidTrie = ()
+
 instance HasTrie () where
   newtype () :->: a = UnitTrie a
   trie f = UnitTrie (f ())
   untrie (UnitTrie a) = \ () -> a
   enumerate (UnitTrie a) = [((),a)]
 
+instance Newtype (() :->: a) where
+  type O (() :->: a) = a
+  pack a = UnitTrie a
+  unpack (UnitTrie a) = a
+
 -- Proofs of inverse properties:
 
 {-
@@ -225,6 +286,11 @@ instance HasTrie Bool where
   untrie (BoolTrie f t) = if' f t
   enumerate (BoolTrie f t) = [(False,f),(True,t)]
 
+instance Newtype (Bool :->: a) where
+  type O (Bool :->: a) = (a,a)
+  pack (a,a') = BoolTrie a a'
+  unpack (BoolTrie a a') = (a,a')
+
 -- | Conditional with boolean last.
 -- Spec: @if' (f False) (f True) == f@
 if' :: x -> x -> Bool -> x
@@ -256,6 +322,11 @@ instance (HasTrie a, HasTrie b) => HasTrie (Either a b) where
   untrie (EitherTrie s t) = either (untrie s) (untrie t)
   enumerate (EitherTrie s t) = enum' Left s `weave` enum' Right t
 
+instance Newtype (Either a b :->: x) where
+  type O (Either a b :->: x) = (a :->: x, b :->: x)
+  pack (f,g) = EitherTrie f g
+  unpack (EitherTrie f g) = (f,g)
+
 enum' :: (HasTrie a) => (a -> a') -> (a :->: b) -> [(a', b)]
 enum' f = (fmap.first) f . enumerate
 
@@ -295,6 +366,11 @@ instance (HasTrie a, HasTrie b) => HasTrie (a,b) where
   enumerate (PairTrie tt) =
     [ ((a,b),x) | (a,t) <- enumerate tt , (b,x) <- enumerate t ]
 
+instance Newtype ((a,b) :->: x) where
+  type O ((a,b) :->: x) = a :->: b :->: x
+  pack abx = PairTrie abx
+  unpack (PairTrie abx) = abx
+
 {-
     untrie (trie f)
       == { trie def }
@@ -563,7 +639,8 @@ f1' = memo f1
 instance HasTrie (V1 x) where
   data (V1 x :->: b) = V1Trie 
   trie _ = V1Trie 
-  untrie V1Trie = \case 
+  untrie V1Trie = \ _ -> error "untrie V1Trie"
+                  -- \case  -- needs EmptyCase
   enumerate V1Trie = [] 
 
 -- | just like @()@ 
@@ -605,6 +682,7 @@ instance (HasTrie (f x)) => HasTrie (M1 i t f x) where
 -- "unlifted" generic representation. (i.e. is a unary type constructor). 
 type Reg a = Rep a () 
 
+-- | 'Generic'-friendly default for 'trie'
 trieGeneric :: (Generic a, HasTrie (Reg a))
             => ((Reg a :->: b) -> (a :->: b))
             -> (a -> b)
@@ -612,6 +690,7 @@ trieGeneric :: (Generic a, HasTrie (Reg a))
 trieGeneric theConstructor f = theConstructor (trie (f . to))
 {-# INLINEABLE trieGeneric #-}
 
+-- | 'Generic'-friendly default for 'untrie'
 untrieGeneric :: (Generic a, HasTrie (Reg a))
               => ((a :->: b) -> (Reg a :->: b))
               -> (a :->: b)
@@ -619,6 +698,7 @@ untrieGeneric :: (Generic a, HasTrie (Reg a))
 untrieGeneric theDestructor t = \a -> (untrie (theDestructor t)) (from a)
 {-# INLINEABLE untrieGeneric #-}
 
+-- | 'Generic'-friendly default for 'enumerate'
 enumerateGeneric :: (Generic a, HasTrie (Reg a))
                  => ((a :->: b) -> (Reg a :->: b))
                  -> (a :->: b)
@@ -643,4 +723,3 @@ dropSum s = case s of
 liftSum :: Either (f a) (g a) -> (f :+: g) a 
 liftSum = either L1 R1
 {-# INLINEABLE liftSum #-}
-
